import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np
from pyH2A.Analysis.Monte_Carlo_Analysis import Monte_Carlo_Analysis
from pyH2A.Utilities.input_modification import convert_input_to_dictionary
from pyH2A.Utilities.output_utilities import insert_image, Figure_Lean

class Comparative_MC_Analysis:
	'''Comparison of Monte Carlo analysis results for different models.

	Parameters
	----------
	Comparative_MC_Analysis > [...] > Value : str
		Path to input file for model.
	Comparative_MC_Analysis > [...] > Image : str, optional
		Path to image for model.

	Notes
	-----
	First column of `Comparative_MC_Analysis` table can include arbitrary 
	name for model.
	'''

	def __init__(self, input_file):
		self.inp = convert_input_to_dictionary(input_file)
		self.models = self.get_models()
		self.check_target_price_range_consistency()

	def get_models(self):
		'''Get models which are to be compared from `Comparative_MC_Analysis` 
		table in input file and perform Monte Carlo analysis for them.
		'''

		models = {}

		for key, input_file in self.inp['Comparative_MC_Analysis'].items():
			model = Monte_Carlo_Analysis(input_file['Value'])

			models[key] = {}
			models[key]['Model'] = model

			try:
				models[key]['Image'] = input_file['Image']
			except KeyError:
				pass

		return models

	def check_target_price_range_consistency(self):
		'''Check that the same target price ranges are specified
		for all models which are to be compared.
		'''

		model_keys = list(self.models)

		target_price_range = np.asarray(self.models[model_keys[0]]['Model'].target_price_range)
		for model_key in model_keys[1:]:
			model_range = self.models[model_key]['Model'].target_price_range
			assert np.array_equal(target_price_range, model_range), 'Target price ranges differ between input files. Reference: {0}, model: {1}'.format(target_price_range, model_range)

		self.target_price_range = target_price_range

	def plot_comparative_distance_histogram(self, ax = None, figure_lean = True, 
											table_kwargs = {}, image_kwargs = {}, 
											plot_kwargs = {}, hist_kwargs = {},
										    **kwargs):
		'''Plot comparative development distance histogram.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		table_kwargs : dict, optional
			Dictionary containing optional keyword arguments for 
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.render_parameter_table`
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		hist_kwargs: dict, optional
			Dictionary containg optional keyword arguments for
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.plot_distance_histogram`
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.
		'''

		model_number = len(self.models)

		kwargs = {**{'left': 0.2, 'right': 0.55, 'bottom': 0.15, 'top': 0.9, 'hspace': 0.2,
	 				 'nrows': model_number, 'ncols': 1,	'sharex': True,	     
	 			     'fig_width': 11, 'fig_height': 1.7 * model_number, 
	 			     'font_size': 12,
	 			     'name': 'Monte_Carlo_Comparative_Distance_Histogram'}, 
	 			  **kwargs, **plot_kwargs}

		table_kwargs = {**{'colWidths': [0.55, 0.25, 0.07, 0.25]}, 
						**table_kwargs}

		image_kwargs = {**{'path': None, 'x': -0.35, 'y': 0.5, 'zoom': 0.08}, 
						**image_kwargs}

		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax

		for counter, (model_name, model) in enumerate(self.models.items()):
			if counter == 0:
				show_title = True
				show_xlabel = False

			elif counter == model_number - 1:
				show_title = False
				show_xlabel = True

			else:
				show_title = False
				show_xlabel = False

			model['Model'].plot_distance_histogram(ax = ax[counter], 
											       figure_lean = False, 
												   xlabel = show_xlabel, 
												   title = show_title,
												   table_kwargs = table_kwargs,
												   **hist_kwargs)

			if 'Image' in model:
				image_kwargs['path'] = model['Image']
				insert_image(ax = ax[counter], **image_kwargs)

		if figure_lean is True:
			figure.execute()
			return figure.fig

	def plot_comparative_distance_cost_relationship(self, ax = None, figure_lean = True,
													table_kwargs = {}, image_kwargs = {}, 
											        plot_kwargs = {}, dist_kwargs = {},
													**kwargs):
		'''Plot comparative development distance/H2 cost relationship.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		table_kwargs : dict, optional
			Dictionary containing optional keyword arguments for 
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.render_parameter_table`
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		dist_kwargs: dict, optional
			Dictionary containg optional keyword arguments for
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.plot_distance_cost_relationship`
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : Figure_Lean object
			Figure_Lean object is returned.
		'''

		model_number = len(self.models)

		kwargs = {**{'right': 0.43, 'left': 0.08, 'top': 0.92,
	 			     'fig_width': 10, 'fig_height': 6, 
	 			     'font_size': 12,
	 			     'name': 'Monte_Carlo_Comparative_Distance_Cost_Relationship'}, 
	 			  **kwargs, **plot_kwargs}

		image_kwargs = {**{'path': None, 'x': 1.2, 'y': 0.5, 'zoom': 0.08}, 
						**image_kwargs}

		dist_kwargs = {**{'log_scale': False, 'legend_loc': 'upper right'},
					   **dist_kwargs}

		table_kwargs = {**{'height': 0.23}, **table_kwargs}

		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax
			return_figure = True
		else:
			return_figure = False

		ax.axhspan(self.target_price_range[0], self.target_price_range[1], color = 'grey', alpha = 0.7)

		for counter, (model_name, model) in enumerate(self.models.items()):
			ycoord = 1 - 1.1 * (counter / model_number) - table_kwargs['height']

			table_kwargs['ypos'] = ycoord

			model['Model'].plot_distance_cost_relationship(ax = ax, figure_lean = False, 
														   table_kwargs = table_kwargs,
														   **dist_kwargs)

			if 'Image' in model:
				image_kwargs['path'] = model['Image']
				image_kwargs['y'] = ycoord + 0.12
				insert_image(ax = ax, **image_kwargs)

		if figure_lean is True:
			figure.execute()
		
		if return_figure:
			return figure

	def plot_combined_distance(self, fig_width = 12, fig_height = 2,
							   table_kwargs = {}, image_kwargs = {}, 
							   plot_kwargs = {}, dist_kwargs = {},
							   hist_kwargs = {}, **kwargs):
		'''Plot combining development distance histogram and distance/H2 cost
		relationship.

		Parameters
		----------
		fig_width : float, optional
			Width of figure in inches.
		fig_height : float, optional
			Height of figure per model in inches.
		target_line : float, optional
			y axis coordinate of target price line.
		table_kwargs : dict, optional
			Dictionary containing optional keyword arguments for 
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.render_parameter_table`
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		dist_kwargs: dict, optional
			Dictionary containg optional keyword arguments for
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.plot_distance_cost_relationship`
		hist_kwargs: dict, optional
			Dictionary containg optional keyword arguments for
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.plot_distance_histogram`
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.
		'''

		model_number = len(self.models)

		kwargs = {**{'right': 0.675, 'left': 0.04, 'bottom': 0.15, 'top': 0.9,
				     'hspace': 0.2, 'wspace': 0.62,
	 			     'name': 'Monte_Carlo_Combined_Plot'}, 
	 			  **kwargs, **plot_kwargs}

		image_kwargs = {**{'x': -0.4}, 
						**image_kwargs}

		dist_kwargs = {**{'log_scale': False, 'legend_loc': 'upper right'},
					   **dist_kwargs}

		gs = gridspec.GridSpec(model_number, 2)

		fig = plt.figure(figsize = (fig_width, fig_height * model_number))
		ax0 = plt.subplot(gs[:, 0])

		ax_hist = []
		for i in range(model_number):
			ax_hist.append(plt.subplot(gs[i, 1]))

		figure = Figure_Lean(provided_figure_and_axis = (fig, ax0), **kwargs)

		for model_name, model in self.models.items():
			model['Model'].plot_distance_cost_relationship(figure_lean = False, ax = ax0 ,
														   parameter_table = False,
														   **dist_kwargs)
		
		self.plot_comparative_distance_histogram(figure_lean = False, ax = ax_hist, 
											     table_kwargs = table_kwargs,
											     image_kwargs = image_kwargs,
											     hist_kwargs = hist_kwargs)

		ax0.axhspan(self.target_price_range[0], self.target_price_range[1], color = 'grey', alpha = 0.7)

		ax0.text(-0.16, 1.04, 'A', transform=ax0.transAxes, size = 24, weight='bold')
		ax0.text(1.36, 1.04, 'B', transform=ax0.transAxes, size = 24, weight='bold')

		figure.execute()

		return fig

	


from pyH2A.Utilities.input_modification import insert, sum_all_tables, process_table, read_textfile
import pyH2A.Utilities.find_nearest as fn
import numpy as np

class Variable_Operating_Cost_Plugin:
	'''Calculation of variable operating costs.

	Parameters
	----------
	Technical Operating Parameters and Specifications > Output per Year > Value : float
		Yearly output taking operating capacity factor into account, in (kg of H2)/year.
	Utilities > [...] > Cost : float, ndarray or str
		Cost of utility (e.g. $/kWh for electricity). May be either a float, a ndarray with the
		same length as `dcf.inflation_correction` or a textfile containing cost values (cost values 
		have to be in second column).
	Utilities > [...] > Usage per kg H2 : float
		Usage of utility per kg H2 (e.g. kWh/(kg of H2) for electricity).
	Utilities > [...] > Price Conversion Factor : float
		Conversion factor between cost and usage units. Should be set to 1 if no conversion is
		required.
	Utilities > [...] > Path : str, optional
		Path for `Cost` entry.
	Utilities > [...] > Usage Path : str, optional
		Path for `Usage per kg H2` entry.
	[...] Other Variable Operating Operating Cost [...] >> Value : float
		``sum_all_tables()`` is used.

	Returns
	-------
	[...] Other Variable Operating Cost [...] > Summed Total > Value : float
		Summed total for each individual table in "Other Variable Operating Cost"
		group.
	Variable Operating Costs > Total > Value : ndarray
		Sum of inflation corrected utilities costs and other variable operating costs.
	Variable Operating Costs > Utilities > Value : ndarray
		Sum of inflation corrected utilities costs.
	Variable Operating Costs > Other > Value : float
		Sum of `Other Variable Operating Cost` entries.
	'''

	def __init__(self, dcf, print_info):
		process_table(dcf.inp, 'Technical Operating Parameters and Specifications', 'Value')
		process_table(dcf.inp, 'Utilities', ['Cost', 'Usage per kg H2'], path_key = ['Path', 'Usage Path'])

		self.calculate_utilities_cost(dcf)
		self.other_variable_costs(dcf, print_info)

		insert(dcf, 'Variable Operating Costs', 'Total', 'Value', 
				self.utilities + self.other, __name__, print_info = print_info)
		insert(dcf, 'Variable Operating Costs', 'Utilities', 'Value', 
				self.utilities, __name__, print_info = print_info)
		insert(dcf, 'Variable Operating Costs', 'Other', 'Value', 
				self.other, __name__, print_info = print_info)

	def calculate_utilities_cost(self, dcf):
		'''Iterating over all utilities and computing summed yearly costs.
		'''

		self.utilities = 0.

		for key in dcf.inp['Utilities']:
			utility = Utility(dcf.inp['Utilities'][key], dcf)
			self.utilities += utility.cost_per_kg_H2

		self.utilities = self.utilities * dcf.inp['Technical Operating Parameters and Specifications']['Output per Year']['Value']

	def other_variable_costs(self, dcf, print_info):
		'''Applying ``sum_all_tables()`` to "Other Variable Operating Cost" group.
		'''

		self.other = dcf.chemical_inflator * sum_all_tables(dcf.inp, 'Other Variable Operating Cost', 'Value', 
																insert_total = True, class_object = dcf, 
																print_info = print_info)

class Utility:
	'''Individual utility objects.

	Methods 
	-------
	calculate_cost_per_kg_H2:
		Calculation of utility cost per kg of H2 with inflation correction.
	'''

	def __init__(self, dictionary, dcf):
		self.calculate_cost_per_kg_H2(dictionary, dcf)

	def calculate_cost_per_kg_H2(self, dictionary, dcf):
		'''Calculation of utility cost per kg of H2 with inflation correction.
		'''
		
		if isinstance(dictionary['Cost'], str):
			prices = read_textfile(dictionary['Cost'], delimiter = '	')
			years_idx = fn.find_nearest(prices, dcf.years)
			prices = prices[years_idx]

			self.cost_per_kg_H2 = prices[:,1] * dcf.inflation_correction * dictionary['Price Conversion Factor'] * dictionary['Usage per kg H2']

		else:
			annual_cost_per_kg_H2 = dcf.inflation_correction * dictionary['Cost'] * dictionary['Usage per kg H2'] * dictionary['Price Conversion Factor']
			self.cost_per_kg_H2 = np.ones(len(dcf.inflation_factor)) * annual_cost_per_kg_H2



from pyH2A.Utilities.input_modification import insert, process_table

class Catalyst_Separation_Plugin:
	'''Calculation of cost for catalyst separation (e.g. via nanofiltration).

	Parameters
	----------
	Water Volume > Volume (liters) > Value : float
		Total water volume in liters.
	Catalyst > Lifetime (years) > Value : float
		Lifetime of catalysts in year before replacement is required.
	Catalyst Separation > Filtration cost ($/m3) > Value : float
		Cost of filtration in $ per m3.

	Returns
	-------
	Other Variable Operating Cost - Catalyst Separation > Catalyst Separation (yearly cost) > Value : float
		Yearly cost of catalyst seperation.
	'''

	def __init__(self, dcf, print_info):
		process_table(dcf.inp, 'Water Volume', 'Value')
		process_table(dcf.inp, 'Catalyst Separation', 'Value')

		self.calculate_yearly_filtration_volume(dcf)
		self.calculate_filtration_cost(dcf)

		insert(dcf, 'Other Variable Operating Cost - Catalyst Separation', 
				'Catalyst Separation (yearly cost)', 'Value', self.yearly_cost,
				__name__, print_info = print_info)

	def calculate_yearly_filtration_volume(self, dcf):
		'''Calculation of water volume that has to be filtered per year.
		'''

		fraction_to_be_filtered_yearly = 1./dcf.inp['Catalyst']['Lifetime (years)']['Value']

		yearly_filtration_volume_liters = dcf.inp['Water Volume']['Volume (liters)']['Value'] * fraction_to_be_filtered_yearly
		self.yearly_filtration_volume_m3 = yearly_filtration_volume_liters/1000.

	def calculate_filtration_cost(self, dcf):
		'''Yearly cost of water filtration to remove catalyst.
		'''

		self.yearly_cost = self.yearly_filtration_volume_m3 * dcf.inp['Catalyst Separation']['Filtration cost ($/m3)']['Value']










import numbers
from functools import lru_cache, reduce
import importlib.resources
from importlib import import_module
from pathlib import Path
import ast
import operator
import numpy as np

def import_plugin(plugin_name, plugin_module):
	'''Importing module.

	Parameters
	----------
	plugin_name : str
		Name of module.
	plugin_module : bool, optional
		Flag to differentiate between plugins and analysis modules. If `True`,
		module is imported from `Plugins.` directory. If `False`, it is
		imported from `Analysis.` directory.

	Returns
	-------
	plugin_class:
		Class from imported module.

	Notes
	-----
	Module `plugin_name` is imported. It is assumed that the module contains 
	a class with the same name as `plugin_name`
	'''

	if plugin_module is True:
		prefix = 'pyH2A.Plugins.'
	else:
		prefix = 'pyH2A.Analysis.'

	plugin = import_module(prefix + plugin_name)
	plugin_class = getattr(plugin, plugin_name)

	return plugin_class

def execute_plugin(plugin_name, plugs_dict, plugin_module = True, 
				   nested_dictionary = False, **kwargs):
	'''Executing module.

	Parameters
	----------
	plugin_name : str
		Name of module.
	plugs_dict : dict
		Dictionary to store class object generated from module.
	plugin_module : bool, optional
		Flag to differentiate between plugins and analysis modules. If `True`,
		module is imported from `Plugins.` directory. If `False`, it is
		imported from `Analysis.` directory.
	nested_dictioanry : bool, optional
		If `True`, a sub dictionary is created in `plugs_dict`, where the
		class object is stored.
	**kwargs:
		Keyword arguments passed to class within module.

	Returns 
	-------
	plugin_object:
		Class objected generated from module.

	Notes
	-----
	Module `plugin_name` is imported. It is assumed that the module contains 
	a class with the same name as `plugin_name`. An instance of this class is created
	using `**kwargs`. The class object is then stored in `plugs_dict`.
	'''

	plugin_class = import_plugin(plugin_name, plugin_module)
	plugin_object = plugin_class(**kwargs)

	if nested_dictionary is True:
		plugs_dict[plugin_name] = {}
		plugs_dict[plugin_name]['Module'] = plugin_object
	else:
		plugs_dict[plugin_name] = plugin_object

	return plugin_object

def check_for_meta_module(key):
	'''Checks if `key` is a meta module that is to be executed.

	Notes
	-----
	Meta module is identified by checking if `key` contains the substring
	'Analysis' and does not contain any of the substrings in `exceptions`.
	'''

	exceptions = ['Parameters', 'Methods', 'Arguments', 'Deactivate']
	indicators = ['Analysis']

	if any(exception in key for exception in exceptions):
		return False
	elif any(indicator in key for indicator in indicators):
		return True
	else:
		return False

def file_import(file_name, mode = 'rb', return_path = False):
	'''Importing package file or file at arbitrary path and returning typing.TextIO
	instance.

	Parameters
	----------
	file_name : str
		Path to file to be read. Can be either a regular path or a path
		of the form `package.subdirectory~file_name` to refer to a file 
		in the pyH2A installation.
	mode : str
		Mode for file read. Can be either `r` or `rb`. In case of `r`, 
		a `typing.TextIO` instance is returned. In case of `rb` a 
		`typing.BinaryIO` instance is returned.

	Returns
	-------
	output : typing.BinaryIO or typing.TextIO instance
		Whether a `typing.BinaryIO` or `typing.TextIO` is returned depends 
		on `mode`.
	'''

	if '~' in file_name:
		package, file = file_name.split('~')
		if 'b' in mode:
			output = importlib.resources.open_binary(package, file)
		else:
			output = importlib.resources.open_text(package, file)

		with importlib.resources.path(package, file) as path:
			output_path = path

	else:
		output_path = Path(file_name)
		output = open(output_path, mode = mode)


	if return_path:
		return output_path
	else:
		return output

@lru_cache(maxsize = None)
def read_textfile(file_name, delimiter, mode = 'rb', **kwargs):
	'''Wrapper for genfromtxt with lru_cache for repeated reads of the same file.

	Parameters
	----------
	file_name : str
		Path to file.
	delimiter : str
		Delimiter used in file.
	**kwargs:
		Keyword arguments passed to ``numpy.genfromtxt()``.

	Returns
	-------
	data : ndarray
		Array containing read data.
	'''

	data = np.genfromtxt(file_import(file_name, mode = mode), delimiter = delimiter, **kwargs)

	return data

def num(s):
	'''Converting string to either an int, float, or, if neither is possible, returning the string.

	Parameters
	----------
	s : str
		String to be converted to a number.

	Returns
	-------
	num : int, float or str
		String converted to int, float or str.

	Notes
	-----
	Input strings can contain commas as thousands seperators, which will be removed if the string 
	is otherwise a valid number (float or int).
	If the input string ends with a "%" sign, it will be converted to a number divided by 100.
	'''

	if s[-1] == '%': #String ends with '%', trying to convert percentage into regular value
		try:
			return num(s[:-1])/100.
		except TypeError:
			return str(s)
	else:
		try:
			return int(s.replace(',', ''))
		except ValueError:
			try:
				return float(s.replace(',', ''))
			except ValueError:
				return str(s)

def convert_file_to_dictionary(file):
	'''Convert provided text file into dictionary. Text file has to follow GitHub 
	flavoured Markdown style.

	Parameters
	----------
	file : typing.TextIO
		typing.TextIO instance of file to be converted.

	Returns
	-------
	inp : dict
		Dictionary containing converted data from file.

	Notes
	-----
	Table format:
	
	# Table A name

	First | Second | ...
	
	--- | --- | ---
	
	Entry A | value 1 | ...
	
	Entry B | value 2 | ...

	# Table B name

	...
	
	The table name is used as `top_key`, the entries within the first column
	are used as `middle_key` and the names of the other columns are used as 
	`bottom key`. E.g. {'Table A name': {'Entry A' : {'Second': 'value 1'}}}
	'''

	inp = {}
	table = False
	header = False

	for line in file:
		
		if line[0] == '#':
			variable_name = line.strip(' #\n')
			inp[variable_name] = {}
			header = False

		if line.strip(' ') == '\n':
			table = False
			header = True

		if line[0] == '-':
			table = True
			header = False

		if header is True and line.strip(' ') != '\n':
			header_entries = line.split('|')

		if table is True and line[0] != '-':
			table_entries = line.split('|')
			inp[variable_name][table_entries[0].strip(' ')] = {}

			for i in zip(header_entries[1:], table_entries[1:]):
				header_entry = i[0].strip(' \n')

				if i[1].strip(' \n') == '':
					table_entry = 'n/a'
				else:
					table_entry = num(i[1].strip(' \n'))

				inp[variable_name][table_entries[0].strip(' ')][header_entry] = table_entry 

	file.close()

	return inp

def convert_dict_to_kwargs_dict(dictionary, middle_key = 'Value'):
	'''Converting dictionary generated by ``convert_file_to_dictionary()``
	to a dictionary that can be used to provide keyword arguments.

	Parameters
	----------
	dictionary : dict
		Dictionary to be converted.
	middle_key : str, optional
		Middle key which is present in input dictionary. This key 
		is removed in the process.

	Returns
	-------
	output : dict
		Dictionary suitable to provide keyword arguments.
	'''

	output = {}

	for key, value in dictionary.items():
		try:
			output[key] = ast.literal_eval(str(value[middle_key]))
		except (ValueError, SyntaxError):
			output[key] = value[middle_key]

	return output

def merge(a, b, path=None, update=True):
	'''Deep merge two dictionaries, with b having priority over a'''

	if path is None: path = []

	for key in b:
		if key in a:
			if isinstance(a[key], dict) and isinstance(b[key], dict):
				merge(a[key], b[key], path + [str(key)])
			elif a[key] == b[key]:
				pass # same leaf value
			elif isinstance(a[key], list) and isinstance(b[key], list):
				for idx, val in enumerate(b[key]):
					a[key][idx] = merge(a[key][idx], b[key][idx], path + [str(key), str(idx)], update=update)
			elif update:
				a[key] = b[key]
			else:
				raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
		
		else:
			a[key] = b[key]
	return a

def convert_input_to_dictionary(file, default = 'pyH2A.Config~Defaults.md', merge_default = True):
	'''Reads provided input file (file) and default file, converting both to dictionaries.
	The dictionaries are merged, with the input file having priority.

	Parameters
	----------
	file : str
		Path to input file.
	default : str, optional
		Path to default file.
	merge_default : bool
		Flag to control if input is merged with default file.

	Returns
	-------
	inp : dict
		Input dictionary.
	'''

	inp_file = convert_file_to_dictionary(file_import(file, mode = 'r'))

	if merge_default is False:
		return inp_file

	else:
		inp_default = convert_file_to_dictionary(file_import(default, mode = 'r'))
		return merge(inp_default, inp_file)

def get_by_path(root, items):
	'''Access a nested object in `root` by item sequence.'''
	return reduce(operator.getitem, items, root)

def set_by_path(root, items, value, value_type = 'value'):
	'''Set a value in a nested object in root by item sequence.

	Notes
	-----
	Existing value is either multiplied by provided one 
	(value_type = factor) or is replaced by provided one.
	In-place replacement, should only be used on deep copy of self.inp dictionary
	'''
	if value_type == 'factor':
		get_by_path(root, items[:-1])[items[-1]] *= value
	else:
		get_by_path(root, items[:-1])[items[-1]] = value

def insert(class_object, top_key, middle_key, bottom_key, value, name, 
		   print_info = True, add_processed = True, insert_path = True):
	'''Insert function used in plugins. 

	Parameters
	----------
	class_object: Discounted_Cash_Flow object
		Dicounted_Cash_Flow object with .inp attribute, which is modified.
	top_key : str
		Top key.
	middle_key : str
		Middle key.
	bottom_key : str
		Bottom key.
	Value : int, float, str or ndarray
		Value inserted at top_key > middle_key > bottom_key position.
	name : str
		Name of plugin performing insertion.
	print_info : bool, optional
		Flag to control if information on action of ``insert()`` is printed.
	add_processed : bool, optional
		Flag to control if 'Processed' key is added.
	insert_path : bool, optional
		Flog to control if 'Path' key is added.

	Notes
	-----
	`inp` attribute of `class_object` is modified by inserting `value` at the position defined
	by `top_key > middle_key > bottom_key`.
	If there already is a value at this position, it will be overwritten. In this case, the 
	'Path' entry will be set to 'None' to avoid issues if value at this position already 
	existed and had a path specified.
	If there is not already a value at this position, it will be created
	`name` is the name of plugin using `insert` for insertion.
	If `print_info` is True, action of `insert` will be printed.
	'''

	try:
		class_object.inp[top_key][middle_key][bottom_key] = value
		if insert_path is True: 
			class_object.inp[top_key][middle_key]['Path'] = 'None' # setting path to "None" to avoid processing
		if print_info is True: print("'{0} > {1} > {2}' is being overwritten by {3}".format(top_key, 
																							middle_key, 
																							bottom_key, 
																							name))

	except KeyError:
		try:
			class_object.inp[top_key][middle_key] = {}
			class_object.inp[top_key][middle_key][bottom_key] = value
			if print_info is True: print("'{0} > {1}' is being created by {2}".format(top_key, 
																					  middle_key, 
																					  name))

		except KeyError:
			class_object.inp[top_key] = {}
			class_object.inp[top_key][middle_key] = {}
			class_object.inp[top_key][middle_key][bottom_key] = value
			if print_info is True: print("'{0}' is being created by {1}".format(top_key, name))

	if add_processed is True:
		class_object.inp[top_key][middle_key]['Processed'] = 'Yes'

def parse_parameter(key, delimiter = '>'):
	'''Provided `key` is split at delimiter(s) and returned as cleaned array'''

	key = str(key)

	path_components = key.split(delimiter)
	output = []

	for i in path_components:
		output.append(i.strip(' '))
		
	return output

def reverse_parameter_to_string(parameter):
	'''Reverts processed parameter list to string.'''

	output = str(parameter[0])

	for item in parameter[1:]:
		output += f' > {item}'

	return output

def parse_parameter_to_array(key, delimiter = '>', dictionary = None, top_key = None, 
							 middle_key = None, bottom_key = None, special_values = [], path = None):
	'''parse_parameter() is applied to `key` string and result is converted to num and
	returned in ndarray 
	
	Parameters
	----------
	key : str
		String convert to array.
	delimiter : str, optional
		Delimiter used in string.
	dictionary : dict, optional
		Dictionary used for lookup.
	top_key : str, optional
		Top key for lookup.
	middle_key : str, optional
		Middle key for lookup.
	bottom_key : str, optional
		Bottom key for lookup.
	special_values : list, optional
		If `key` contains an element of `special_values`, the value at `path`
		is retrieved instead of using the actual value of `key`.
	path : str, optional
		Path for lookup in case `special_values` is triggered.

	Returns
	-------
	array : ndarray
		Output as array.
	'''

	output = parse_parameter(key, delimiter = delimiter)
	array = []

	for i in output:
		if i in special_values:
			array.append(process_cell(dictionary, top_key, middle_key, bottom_key, 
						 cell = path, print_processing_warning = False))
		else:
			array.append(process_cell(dictionary, top_key, middle_key, bottom_key, 
						 cell = i, print_processing_warning = False))

	return np.asarray(array)

def process_path(dictionary, path, top_key, key, bottom_key, print_processing_warning = True):
	'''Processing provided path. Checks are performed to see if path is valid.

	Parameters
	----------
	dictionary : dict
		Dictionary within which function operates.
	path : str
		Path.
	top_key : str
		Top key.
	key : str
		Middle key.
	bottom_key : str
		Bottom key.
	print_processing_warning : bool
		Flag to control if a warning is printed when an unprocessed value is being used.

	Notes
	-----
	If provided path contains no ">" symbols, it is not a path and 1 is returned
	If provided path contains only one ">" symbol, it is not a valid path. A warning is printed and 1 is returned.
	If provided path contains two ">" symbols, it is potentially a valid path. 
	It is then attempted to retrieve target value. If retrieval attempt is unsuccessful, a warning is printed and 1 is returned.
	If the path is valid, the target value is retrieved:
	If the rerieved target value comes from an unprocessed key, a warning is printed.
	If the retrieved target value is non-numerical, a warning is printed and 1 is returned.
	If the retrieved target value is numerical, it is returned.
	'''

	parsed_path = parse_parameter(path)

	if len(parsed_path) == 1:
		return 1.

	elif len(parsed_path) == 3:

		try:
			target_value = get_by_path(dictionary, parsed_path)

			if 'Processed' not in dictionary[parsed_path[0]][parsed_path[1]] and print_processing_warning is True:
				print('Warning: Unprocessed value is being used at "{0} > {1} > {2}" (by "{3} > {4}")'
					  .format(parsed_path[0], parsed_path[1], parsed_path[2], top_key, key))

			if not isinstance(target_value, numbers.Number):
				if isinstance(target_value, list) or type(target_value).__module__ == np.__name__:
					pass
				else:
					print('Warning: Non-numerical value retrieved at "{0} > {1} > {2}" (by "{3} > {4}"), setting to 1'
						  .format(parsed_path[0], parsed_path[1], parsed_path[2], top_key, key))
					target_value = 1.

		except KeyError:
			print('Warning: Invalid path specified for "{0}" (at "{1} > {2} > {3}"), setting to 1'
				  .format(path, top_key, key, bottom_key))
			target_value = 1.

		return target_value

	else:
		print('Warning: Invalid path specified for "{0}" (at "{1} > {2} > {3}"), setting to 1'
			   .format(path, top_key, key, bottom_key))
		return 1.

def process_cell(dictionary, top_key, key, bottom_key, cell = None, print_processing_warning = True):
	'''Processing of a single cell at dictionary[top_key][key][bottom_key]

	Parameters
	----------
	dictionary : dict
		Dictionary within which function operates.
	top_key : str
		Top key.
	key : str
		Middle key.
	bottom_key : str
		Bottom key.
	cell : int, float, str or None
		Cell entry.
	print_processing_warning : bool
		Flag to control if a warning is printed when an unprocessed value is being used.

	Notes
	-----
	If cell contains only a number, the contents of that cell are returned. 
	If cell contains a string, but that string is not a path (indicated by absence of ">" symbol), 1 is returned
	If cell contains a string which is potentially a path, it is processed:
	Contents of the cell are split at ";" delimiter, separating multiple potential paths.
	For each potential path, process_path() is applied.
	The retrieved target value(s) are multiplied and returned.
	Since value is initated to 1, if none of the paths are valid, simply 1 is returned.
	'''
	if cell is None:
		cell = dictionary[top_key][key][bottom_key]

	if isinstance(cell, numbers.Number):
		return cell

	elif '>' not in cell:
		if isinstance(num(cell), numbers.Number):
			return num(cell)
		else:
			return 1.

	else:
		value = 1.
		paths = parse_parameter(cell, delimiter = ';')

		for path in paths:
			target_value = process_path(dictionary, path, top_key, key, bottom_key, 
								print_processing_warning = print_processing_warning)
			value *= target_value

		return value

def process_input(dictionary, top_key, key, bottom_key, path_key = 'Path', add_processed = True):
	'''Processing of input at dictionary[top_key][key][bottom_key].

	Parameters
	----------
	dictionary : dict
		Dictionary within which function operates.
	top_key : str
		Top key.
	key : str
		Middle key.
	bottom_key : str
		Bottom key.
	path_key : str, optional
		Key used for path column. Defaults to 'Path'.
	add_processed : bool, optional
		Flag to control if `Processed` key is added

	Notes
	-----
	Action: if there is an entry at dictionary[top_key][key][path_key], process_input() applies process_cell() 
	to dictionary[top_key][key][bottom_key] as well as dictionary[top_key][key][path_key] and multiplies them. 
	The resulting value is returned and placed into dictionary[top_key][key][bottom_key]

	Detailed Description:

	First, it is checked if that input has already been processed by looking for the "Processed" key.
	If this is the case, the input is simply returned.
	If it has not already been processed, it is checked if the input is a string which could not 
	be a path (not containing ">"). In this case the string is simply returned and "Processed" is added.

	If neither condition is met, process_cell() is applied. 
	It is then attempted to retrieve dictionary[top_key][key][path_key]. 
	If this entry cannot be retrieved, the process_cell() value of the input is returned,

	If this entry can be retrieved, process_cell() is applied to it and the resulting
	Value is multiplied by the original process_cell() value of the input, updating value.

	If the obtained value differs from the original entry, the obtained value is inserted at
	dictionary[top_key][key][bottom_key] and the original entry is stored in
	dictionary[top_key][key][former_bottom_key] 

	At the end, "Processed" is added.
	'''

	entry = dictionary[top_key][key][bottom_key]

	if 'Processed' in dictionary[top_key][key]:
		return entry

	elif isinstance(entry, str) and '>' not in entry:
		if add_processed is True:
			dictionary[top_key][key]['Processed'] = 'Yes'
		return entry

	else:
		value = process_cell(dictionary, top_key, key, bottom_key)

		try:
			target_value = process_cell(dictionary, top_key, key, path_key)
			value *= target_value
		except KeyError:
			pass

		if np.array_equal(value, dictionary[top_key][key][bottom_key]) is False:
			former_bottom_key = 'Former ' + bottom_key 
			dictionary[top_key][key][former_bottom_key] = dictionary[top_key][key][bottom_key]
			dictionary[top_key][key][bottom_key] = value  # setting dictionary entry to obtained value
		
		if add_processed is True:
			dictionary[top_key][key]['Processed'] = 'Yes'   # marking that this key has been processed

		return value

def process_table(dictionary, top_key, bottom_key, path_key = 'Path'):
	'''Looping through all keys in dictionary[top_key] and applying process_input to
	dictionary[top_key][key][bottom_key].

	Parameters
	----------
	dictionary : dict
		Dictionary within which function operates.
	top_key : str
		Top key.
	bottom_key : str, ndarray or list
		Bottom key(s).
	path_key : str or ndarray, optional
		Key(s) used for path column(s). Defaults to 'Path'.
	
	Notes
	-----
	`bottom_key` can also be an array of keys, all of which are processed (in this case, 
	path_key has to be an array of equal length). 
	'''

	for key in dictionary[top_key]:
		if isinstance(bottom_key, str):
			value = process_input(dictionary, top_key, key, bottom_key, path_key = path_key)

		else:
			for single_key, path in zip(bottom_key[:-1], path_key[:-1]):
				value = process_input(dictionary, top_key, key, single_key, 
									  path_key = path, add_processed = False)

			process_input(dictionary, top_key, key, bottom_key[-1], 
						  path_key = path_key[-1], add_processed = True)

def sum_table(dictionary, top_key, bottom_key, path_key = 'Path'):
	'''For the provided `dictionary`, all entries in dictionary[top_key] are processed 
	using ``process_input()`` (positions: top_key > key > bottom key) and summed.

	Parameters
	----------
	dictionary : dict
		Dictionary within which function operates.
	top_key : str
		Top key.
	bottom_key : str, ndarray or list
		Bottom key.
	path_key : str, optional
		Key used for path column. Defaults to 'Path'.
	'''

	value = 0.

	for key in dictionary[top_key]:
		value += process_input(dictionary, top_key, key, bottom_key, path_key = path_key)

	return value

def sum_all_tables(dictionary, table_group, bottom_key, insert_total = False, 
				   class_object = None, middle_key_insertion = 'Summed Total', 
				   bottom_key_insertion = 'Value', print_info = True, 
				   path_key = 'Path', return_contributions = False):
	'''Applies ``sum_table()`` to all dictionary entries with a key that contains `table_group`. 
	Resulting ``sum_table()`` values are summed to return total.

	Parameters
	----------
	dictionary : dict
		Dictionary within which function operates.
	table_group : str
		String to identify table group. If a dictionary key contains the `table_group`
		substring it is part of the table group.
	bottom_key : str, ndarray or list
		Bottom key.
	insert_total : bool, optional
		If `insert_total` is True, the total of each table is inserted in the
		respective table.
	class_object : Discounted_Cash_Flow object
		Discounted_Cash_Flow object whose .inp attribute is modified.
	middle_key_insertion : str, optional
		Middle key used for insertion of total.
	bottom_key_insertion : str, optional
		Bottom key used for insertion of total.
	print_info : bool, optional
		Flag to control if information on action of ``insert()`` is printed.
	path_key : str, optional
		Key used for path column. Defaults to 'Path'.
	return_contributions : bool, optional
		Flag to control if a dictionary with contributions breakdown (for use 
		in cost ``Cost_Contributions_Analysis`` module) is returned.

	Notes
	-----
	If `insert_total` is true, the ``sum_table()`` value for a given key is inserted 
	in class_object.inp at key > middle_key_insertion > bottom_key_insertion.

	The contributions of each table in table_group are stored in `contributions` dictionary, 
	which is returned if `return_contributions` is set to True. Dictionary is structured so 
	that it can be provided to "Cost_Contributions_Analysis" class to generate a cost breakdown plot.
	'''

	total = 0.
	contributions = {}
	contributions['Data'] = {}

	for key in dictionary:

		if table_group in key:
			value = sum_table(dictionary, key, bottom_key, path_key = path_key)
			total += value
			contributions['Data'][key] = value

			if insert_total is True:
				insert(class_object, key, middle_key_insertion, bottom_key_insertion, 
					    value, __name__, print_info = print_info)

	contributions['Total'] = total
	contributions['Table Group'] = table_group

	if return_contributions is True:
		return total, contributions
	else:
		return total


from pyH2A.Utilities.input_modification import insert, process_table, read_textfile
import numpy as np

class Photovoltaic_Plugin:
	'''Simulation of hydrogen production using PV + electrolysis.

	Parameters
	----------
	Financial Input Values > construction time > Value : int
		Construction time of hydrogen production plant in years.
	Irradiation Used > Data > Value : str or ndarray
		Hourly power ratio data for electricity production calculation. Either a 
		path to a text file containing the data or ndarray. A suitable array 
		can be retrieved from "Hourly Irradiation > *type of tracking* > Value".
	CAPEX Multiplier > Multiplier > Value : float
		Multiplier to describe cost reduction of PV and electrolysis CAPEX for every ten-fold
		increase of power relative to CAPEX reference power. Based on the multiplier the CAPEX
		scaling factor is calculated as: multiplier ^ (number of ten-fold increases). A value
		of 1 leads to no CAPEX reduction, a value < 1 enables cost reduction.
	Electrolyzer > Nominal Power (kW) > Value : float
		Nominal power of electrolyzer in kW.
	Electrolyzer > CAPEX Reference Power (kW) > Value : float
		Reference power of electrolyzer in kW for cost reduction calculation.
	Electrolyzer > Power requirement increase per year > Value : float
		Electrolyzer power requirement increase per year due to stack degradation. Percentage 
		or value > 0. Increase calculated as: (1 + increase per year) ^ year.
	Electrolyzer > Minimum capacity > Value : float
		Minimum capacity required for electrolyzer operation. Percentage or value between 0 and 1.
	Electrolyzer > Conversion efficiency (kg H2/kWh) > Value : float
		Electrical conversion efficiency of electrolyzer in (kg H2)/kWh.
	Electrolyzer > Replacement time (h) > Value : float
		Operating time in hours before stack replacement of electrolyzer is required.
	Photovoltaic > Nominal Power (kW) > Value : float
		Nominal power of PV array in kW.
	Photovoltaic > CAPEX Reference Power (kW) > Value : float
		Reference power of PV array for cost reduction calculations.
	Photovoltaic > Power loss per year > Value : float
		Reduction in power produced by PV array per year due to degradation. Percentage or value
		> 0. Reduction calculated as: (1 - loss per year) ^ year.
	Photovoltaic > Efficiency > Value : float
		Power conversion efficiency of used solar cells. Percentage or value between 0 and 1.

	Returns
	-------
	Technical Operating Parameters and Specifications > Plant Design Capacity (kg of H2/day) > Value : float
		Plant design capacity in (kg of H2)/day calculated from installed 
		PV + electrolysis power capacity and hourly irradiation data.
	Technical Operating Parameters and Specifications >	Operating Capacity Factor (%) > Value : float
		Operating capacity factor is set to 1 (100%).
	Planned Replacement > Electrolyzer Stack Replacement > Frequency (years) : float
		Frequency of electrolyzer stack replacements in years, calculated from replacement time and hourly
		irradiation data.
	Electrolyzer > Scaling Factor > Value : float
		CAPEX scaling factor for electrolyzer calculated based on CAPEX multiplier, 
		reference and nominal power.
	Photovoltaic > Scaling Factor > Value : float
		CAPEX scaling factor for PV array calculated based on CAPEX multiplier, 
		reference and nominal power.
	Non-Depreciable Capital Costs > Land required (acres) > Value : float
		Total land required in acres.
	Non-Depreciable Capital Costs > Solar Collection Area (m2) > Value : float
		Solar collection area in m2.
	'''

	def __init__(self, dcf, print_info):
		process_table(dcf.inp, 'Irradiation Used', 'Value')
		process_table(dcf.inp, 'CAPEX Multiplier', 'Value')
		process_table(dcf.inp, 'Electrolyzer', 'Value')
		process_table(dcf.inp, 'Photovoltaic', 'Value')

		self.calculate_H2_production(dcf)
		self.calculate_stack_replacement(dcf)
		self.calculate_scaling_factors(dcf)
		self.calculate_area(dcf)

		insert(dcf, 'Technical Operating Parameters and Specifications', 'Plant Design Capacity (kg of H2/day)', 'Value', 
			   self.h2_production/365., __name__, print_info = print_info)
		insert(dcf, 'Technical Operating Parameters and Specifications', 'Operating Capacity Factor (%)', 'Value', 
				1., __name__, print_info = print_info)
	
		insert(dcf, 'Planned Replacement', 'Electrolyzer Stack Replacement', 'Frequency (years)', 
				self.replacement_frequency, __name__, print_info = print_info, add_processed = False,
				insert_path = False)

		insert(dcf, 'Electrolyzer', 'Scaling Factor', 'Value', 
				self.electrolyzer_scaling_factor, __name__, print_info = print_info)
		insert(dcf, 'Photovoltaic', 'Scaling Factor', 'Value', 
				self.pv_scaling_factor, __name__, print_info = print_info)

		insert(dcf, 'Non-Depreciable Capital Costs', 'Land required (acres)', 'Value', 
				self.area_acres, __name__, print_info = print_info)
		insert(dcf, 'Non-Depreciable Capital Costs', 'Solar Collection Area (m2)', 'Value', 
				self.area_m2, __name__, print_info = print_info)

	def calculate_H2_production(self, dcf):
		'''Using hourly irradiation data and electrolyzer as well as PV array parameters,
		H2 production is calculated.
		'''

		if isinstance(dcf.inp['Irradiation Used']['Data']['Value'], str):
			data = read_textfile(dcf.inp['Irradiation Used']['Data']['Value'], delimiter = '	')[:,1]
		else:
			data = dcf.inp['Irradiation Used']['Data']['Value']

		yearly_data = []

		for year in dcf.operation_years:
			data_loss_corrected = self.calculate_photovoltaic_loss_correction(dcf, data, year)
			power_generation = data_loss_corrected * dcf.inp['Photovoltaic']['Nominal Power (kW)']['Value']

			electrolyzer_power_demand, power_increase = self.calculate_electrolyzer_power_demand(dcf, year) 
			electrolyzer_power_demand *= np.ones(len(power_generation))
			electrolyzer_power_consumption = np.amin(np.c_[power_generation, electrolyzer_power_demand], axis = 1)

			threshold = dcf.inp['Electrolyzer']['Minimum capacity']['Value']
			electrolyzer_capacity = electrolyzer_power_consumption / electrolyzer_power_demand
			electrolyzer_capacity[electrolyzer_capacity > threshold] = 1
			electrolyzer_capacity[electrolyzer_capacity <= threshold] = 0

			h2_produced = electrolyzer_power_consumption * dcf.inp['Electrolyzer']['Conversion efficiency (kg H2/kWh)']['Value'] / power_increase
			h2_produced *= electrolyzer_capacity

			yearly_data.append([year, np.sum(h2_produced), np.sum(electrolyzer_capacity)])

		self.yearly_data = np.asarray(yearly_data)
		self.h2_production = np.concatenate([np.zeros(dcf.inp['Financial Input Values']['construction time']['Value']), 
												self.yearly_data[:,1]])

	def calculate_photovoltaic_loss_correction(self, dcf, data, year):
		'''Calculation of yearly reduction in electricity production by PV array.
		'''

		return data * (1. - dcf.inp['Photovoltaic']['Power loss per year']['Value']) ** year

	def calculate_electrolyzer_power_demand(self, dcf, year):
		'''Calculation of yearly increase in electrolyzer power demand.
		'''

		increase = (1. + dcf.inp['Electrolyzer']['Power requirement increase per year']['Value']) ** year
		demand = increase * dcf.inp['Electrolyzer']['Nominal Power (kW)']['Value']

		return demand, increase

	def calculate_stack_replacement(self, dcf):
		'''Calculation of stack replacement frequency for electrolyzer.
		'''

		cumulative_running_time = np.cumsum(self.yearly_data[:,2])
		stack_usage = cumulative_running_time / dcf.inp['Electrolyzer']['Replacement time (h)']['Value']

		number_of_replacements = np.floor_divide(stack_usage[-1], 1)

		self.replacement_frequency = len(stack_usage) / (number_of_replacements + 1.)

	def calculate_scaling_factors(self, dcf):
		'''Calculation of electrolyzer and PV CAPEX scaling factors.
		'''

		self.pv_scaling_factor = self.scaling_factor(dcf, dcf.inp['Photovoltaic']['Nominal Power (kW)']['Value'], dcf.inp['Photovoltaic']['CAPEX Reference Power (kW)']['Value'])
		self.electrolyzer_scaling_factor = self.scaling_factor(dcf, dcf.inp['Electrolyzer']['Nominal Power (kW)']['Value'], dcf.inp['Electrolyzer']['CAPEX Reference Power (kW)']['Value'])
		
	def scaling_factor(self, dcf, power, reference):
		'''Calculation of CAPEX scaling factor based on nominal and reference power.
		'''
		
		number_of_tenfold_increases = np.log10(power/reference)

		return dcf.inp['CAPEX Multiplier']['Multiplier']['Value'] ** number_of_tenfold_increases

	def calculate_area(self, dcf):
		'''Area requirement calculation assuming 1000 W/m2 peak power.'''

		peak_kW_per_m2 = dcf.inp['Photovoltaic']['Efficiency']['Value'] * 1.
		self.area_m2 = dcf.inp['Photovoltaic']['Nominal Power (kW)']['Value'] / peak_kW_per_m2
		self.area_acres = self.area_m2 * 0.000247105




import copy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from pyH2A.Discounted_Cash_Flow import Discounted_Cash_Flow
from pyH2A.Utilities.input_modification import num, convert_input_to_dictionary, parse_parameter, get_by_path, set_by_path
from pyH2A.Utilities.output_utilities import make_bold, Figure_Lean, dynamic_value_formatting

import pprint

class Sensitivity_Analysis:
	'''Sensitivity analysis for multiple parameters.

	Parameters
	----------
	Sensitivity_Analysis > [...] > Name : str
		Display name for parameter, e.g. used for plot labels.
	Sensitivity_Analysis > [...] > Type : str
		Type of parameter values. If `Type` is 'value', provided values are
		used as is. If `Type` is 'factor', provided values are multiplied 
		with base value of parameter in input file.
	Sensitivity_Analysis > [...] > Values : str
		Value pair to be used for sensitivity analysis. One value should 
		be higher than the base value, the other should be lower.
		Specified in following format: value A; value B (order is irrelevant).
		E.g. '0.3; 10'.

	Notes
	-----
	`Sensitivity_Analysis` contains parameters which are to be varied in
	sensitivity analysis. First column specifies path to parameter in input file 
	(top key > middle key > bottom key format, e.g. Catalyst > Cost per kg ($) > Value).
	Order of parameters is not relevant.
	'''

	def __init__(self, input_file):
		self.inp = convert_input_to_dictionary(input_file)
		self.base_case = Discounted_Cash_Flow(input_file, print_info = False)
		#self.results = self.perform_sensitivity_analysis()

	def perform_sensitivity_analysis(self, format_cutoff = 7):
		'''Perform sensitivity analysis.

		Parameters
		----------
		format_cutoff : int
			Length of number string above which dyanmic formatting is applied.
		'''

		sensitivity_results = {}

		for key in self.inp['Sensitivity_Analysis']:
			parameters = parse_parameter(key)
			name = self.inp['Sensitivity_Analysis'][key]['Name']



			sensitivity_results[name] = {}
			sensitivity_results[name]['Base'] = dynamic_value_formatting(get_by_path(self.inp, parameters),
																		 cutoff = format_cutoff)
			sensitivity_results[name]['Values'] = {}

			values = parse_parameter(self.inp['Sensitivity_Analysis'][key]['Values'], 
									 delimiter = ';')

			for value in values:
				input_dict = copy.deepcopy(self.inp)
				numerical_value = num(value)

				value_type = self.inp['Sensitivity_Analysis'][key]['Type']

				set_by_path(input_dict, parameters, numerical_value, 
							value_type = value_type)

				if self.inp['Sensitivity_Analysis'][key]['Type'] == 'factor':
					sensitivity_results[name]['Base'] = '1.0x'
					shown_value = '{0}x'.format(numerical_value)
				else:
					
					if '%' in value:
						shown_value = value
						sensitivity_results[name]['Base'] = '{0}%'.format(dynamic_value_formatting(
																		   get_by_path(self.inp, 
																					   parameters) * 100),
																			cutoff = format_cutoff)
					else:
						shown_value = dynamic_value_formatting(numerical_value, cutoff = format_cutoff)


				dcf = Discounted_Cash_Flow(input_dict, print_info = False)

				sensitivity_results[name]['Values'][shown_value] = dcf.h2_cost

		return sensitivity_results

	def sort_h2_cost_values(self, data):
		'''Sort H2 cost values.
		'''

		for key in data:
			values = data[key]['Values']

			low_key = min(values, key = values.get)
			high_key = max(values, key = values.get)

			data[key]['Low - Name'] = low_key
			data[key]['Low - Value'] = values[low_key]

			data[key]['High - Name'] = high_key
			data[key]['High - Value'] = values[high_key]

			del data[key]['Values']

	def sensitivity_box_plot(self, ax = None, figure_lean = True,
							 height = 0.8, label_offset = 0.1, 
						     lim_extra = 0.2, format_cutoff = 7,
						     plot_kwargs = {}, **kwargs):
		'''Plot sensitivity box plot.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		height : float, optional
			Height of bars.
		label_offset : float, optional
			Offset for bar labels.
		lim_extra : float, optional
			Fractional increase of x axis limits.
		format_cutoff : int
			Length of number string above which dyanmic formatting is applied.
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.

		Notes
		-----
		In plot, parameters are sorted by descending cost increase magnitude.
		'''

		self.results = self.perform_sensitivity_analysis(format_cutoff = format_cutoff)		

		data = copy.deepcopy(self.results)

		self.sort_h2_cost_values(data)

		df = pd.DataFrame.from_dict(data)
		df.sort_values(by = ['High - Value'], axis = 1, 
					   na_position = 'first', inplace = True)

		kwargs = {**{'left': 0.45, 'right': 0.99, 'bottom': 0.2, 'top': 0.95,
				     'fig_width': 7.5, 'fig_height': 4,
				     'name': 'Sensitivity_Box_Plot'}, 
				  **kwargs, **plot_kwargs}
		
		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax		

		number_of_entries = len(df.columns)
		base_case = self.base_case.h2_cost

		max_value = df.loc['High - Value'].max(skipna = True)
		min_value = df.loc['Low - Value'].min(skipna = True)

		x_width = max_value - min_value
		label_offset = x_width * label_offset
		extra = x_width * lim_extra

		xlim = [min_value - extra, max_value + extra]
		ylim = [-height, number_of_entries]	

		labels = []

		for counter, key in enumerate(df):
			lower = df[key]['Low - Value']
			upper = df[key]['High - Value']

			rectangle_left = patches.Rectangle((lower, counter-height/2), 
												base_case - lower, height, 
												edgecolor = 'none', 
												facecolor = 'darkgreen')
			rectangle_right = patches.Rectangle((base_case, counter-height/2), 
												 upper - base_case, height, 
												 edgecolor = 'none', 
												 facecolor = 'darkred')

			ax.add_patch(rectangle_left)
			ax.add_patch(rectangle_right)

			ax.annotate('${:.2f}'.format(lower), xy = (lower - label_offset, counter), 
				         va = 'center', ha = 'center')
			ax.annotate('${:.2f}'.format(upper), xy = (upper + label_offset, counter), 
				         va = 'center', ha = 'center')

			labels.append('{0}\n{1}, {2}, {3}'.format(make_bold(key), 
				                                      df[key]['Low - Name'], 
				                                      df[key]['Base'], 
				                                      df[key]['High - Name']))

		ax.set_xlabel(r'Cost sensitivity / USD per kg $H_{2}$')
		ax.set_yticks(np.arange(0, number_of_entries))
		ax.set_yticklabels(labels)
		ax.grid(color = 'grey', linestyle = '--', linewidth = 0.2)

		ax.set_xlim(xlim[0], xlim[1])
		ax.set_ylim(ylim[0], ylim[1])
		ax.plot((base_case, base_case), (ylim[0], ylim[1]), '--', color = 'black')

		if figure_lean is True:
			figure.execute()
			return figure.fig



from pyH2A.Utilities.input_modification import insert, sum_all_tables, process_table

class Capital_Cost_Plugin:
	'''

	Parameters
	----------
	[...] Direct Capital Cost [...] >> Value : float
		``sum_all_tables()`` is used.
	[...] Indirect Capital Cost [...] >> Value : float
		``sum_all_tables()`` is used.
	Non-Depreciable Capital Costs > Cost of land ($ per acre) > Value : float
		Cost of land in $ per acre, ``process_table()`` is used.
	Non-Depreciable Capital Costs > Land required (acres) > Value : float
		Total land are required in acres, ``process_table()`` is used.
	[...] Other Non-Depreciable Capital Cost [...] >> Value : float
		``sum_all_tables()`` is used.

	Returns
	-------
	[...] Direct Capital Cost [...] > Summed Total > Value : float
		Summed total for each individual table in "Direct Capital Cost" group.
	[...] Indirect Capital Cost [...] > Summed Total > Value : float
		Summed total for each individual table in "Indirect Capital Cost" group.
	[...] Other Non-Depreciable Capital Cost  [...] > Summed Total > Value : float
		Summed total for each individual table in "Other Non-Depreciable Capital Cost" group.
	Direct Capital Costs > Total > Value : float
		Total direct capital costs.
	Direct Capital Costs > Inflated > Value : float
		Total direct capital costs multiplied by combined inflator.
	Indirect Capital Costs > Total > Value : float
		Total indirect capital costs.
	Indirect Capital Costs > Inflated > Value : float
		Total indirect capital costs multiplied by combined inflator.
	Non-Depreciable Capital Costs > Total > Value : float
		Total non-depreciable capital costs.
	Non-Depreciable Capital Costs > Inflated > Value : float
		Total non-depreciable capital costs multiplied by combined inflator.
	Depreciable Capital Costs > Total > Value : float
		Sum of direct and indirect capital costs.
	Depreciable Capital Costs > Inflated > Value : float
		Sum of direct and indirect capital costs multiplied by combined inflator.
	Total Capital Costs > Total > Value : float
		Sum of depreciable and non-depreciable capital costs.
	Total Capital Costs > Inflated > Value : float
		Sum of depreicable and non-depreciable capital costs multiplied by combined inflator.
	['Capital_Cost_Plugin'].direct_contributions : dict
		Attribute containing cost contributions for "Direct Capital Cost" group.
	'''
	def __init__(self, dcf, print_info):

		self.direct_capital_costs(dcf, print_info)  
		direct_inflated = self.direct * dcf.combined_inflator

		insert(dcf, 'Direct Capital Costs', 'Total', 'Value', self.direct, __name__, print_info = print_info)
		insert(dcf, 'Direct Capital Costs', 'Inflated', 'Value', direct_inflated, __name__, print_info = print_info)

		self.indirect_capital_costs(dcf, print_info)

		indirect_inflated = self.indirect * dcf.combined_inflator
		depreciable = self.direct + self.indirect
		depreciable_inflated = direct_inflated + indirect_inflated
		
		insert(dcf, 'Indirect Capital Costs', 'Total', 'Value', self.indirect, __name__, print_info = print_info)
		insert(dcf, 'Indirect Capital Costs', 'Inflated', 'Value', indirect_inflated, __name__, print_info = print_info)
		
		insert(dcf, 'Depreciable Capital Costs', 'Total', 'Value', depreciable, __name__, print_info = print_info)
		insert(dcf, 'Depreciable Capital Costs', 'Inflated', 'Value', depreciable_inflated, __name__, print_info = print_info)
		
		self.non_depreciable_capital_costs(dcf, print_info)

		non_depreciable_inflated = self.non_depreciable * dcf.ci_inflator
		total = depreciable + self.non_depreciable
		total_inflated = depreciable_inflated + non_depreciable_inflated

		insert(dcf, 'Non-Depreciable Capital Costs', 'Total', 'Value', self.non_depreciable, __name__, print_info = print_info)
		insert(dcf, 'Non-Depreciable Capital Costs', 'Inflated', 'Value', non_depreciable_inflated, __name__, print_info = print_info)
		
		insert(dcf, 'Total Capital Costs', 'Total', 'Value', total, __name__, print_info = print_info)
		insert(dcf, 'Total Capital Costs', 'Inflated', 'Value', total_inflated, __name__, print_info = print_info)

	def direct_capital_costs(self, dcf, print_info):
		'''Calculation of direct capital costs by applying ``sum_all_tables()`` to "Direct Capital Cost" group.'''

		self.direct, self.direct_contributions = sum_all_tables(dcf.inp, 'Direct Capital Cost', 'Value', insert_total = True, 
																class_object = dcf, print_info = print_info, 
																return_contributions = True)

	def indirect_capital_costs(self, dcf, print_info):
		'''Calculation of indirect capital costs by applying ``sum_all_tables()`` to "Indirect Capital Cost" group.'''

		self.indirect = sum_all_tables(dcf.inp, 'Indirect Capital Cost', 'Value', insert_total = True, 
									   class_object = dcf, print_info = print_info)

	def non_depreciable_capital_costs(self, dcf, print_info):
		'''Calculation of non-depreciable capital costs by calculating cost of land and applying
		``sum_all_tables()`` to "Other Non-Depreciable Capital Cost" group.
		'''

		process_table(dcf.inp, 'Non-Depreciable Capital Costs', 'Value')

		non_depreciable = dcf.inp['Non-Depreciable Capital Costs']
		self.non_depreciable = non_depreciable['Cost of land ($ per acre)']['Value'] * non_depreciable['Land required (acres)']['Value']
		self.non_depreciable += sum_all_tables(dcf.inp, 'Other Non-Depreciable Capital Cost', 'Value', insert_total = True, class_object = dcf, print_info = print_info)




import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
from scipy.optimize import least_squares

from pyH2A.Analysis.Monte_Carlo_Analysis import Monte_Carlo_Analysis, calculate_distance
from pyH2A.Utilities.find_nearest import find_nearest
from pyH2A.Utilities.input_modification import convert_input_to_dictionary, read_textfile, file_import, parse_parameter, merge
from pyH2A.Utilities.output_utilities import Figure_Lean, insert_image

def linear(p, x):
	'''Linear function for fitting.
	'''

	return p[0] * x + p[1]

def exponential_decline(p, x):
	'''Exponential decay function for fitting.
	'''

	return p[0] * np.exp(-p[1] * x)

def residual_generic(p, x, y, function, **kwargs):
	'''Generic residual function for least squares fitting.
	'''

	y_fit = function(p, x, **kwargs)
	res = y - y_fit

	return res

def fit_generic(x, y, function, p_guess = None, kwargs = {}):
	'''Generic least squares fitting function.
	'''

	if p_guess is None:
		p_guess = np.ones(2)

	p = least_squares(fun=residual_generic, x0=p_guess, args=(x, y, function), kwargs = kwargs)

	return p.x

class Development_Distance_Time_Analysis:
	'''Relating development distance to time using historical data.

	Parameters
	----------
	Development_Distance_Time_Analysis > Input File > Value : str
		Path to textfile containing historical data for parameters
		specified in Monte Carlo analysis (see Notes).
	Development_Distance_Time_Analysis > Log Normalization > Value : bool
		Boolean flag to control if logarithmic normalization is used
		for distance calculation or not. If `False`, linear normalization is used.
		When analyzing historical data, it is advised to use logaritthmic 
		normalization to avoid outsized impacts of parameters which have changed on
		the order of magnitudes.
	Development_Distance_Time_Analysis > Base Year > Value : int
		Year which corresponds to the `Base` values specified in Monte Carlo analysis.
	Development_Distance_Time_Analysis > Extrapolation Limit > Year : int
		Year until which the distance/time models should be extrapolated.

	Notes
	-----
	The loaded textfile has to contain datapoints with historical parameter values for 
	the specified technology.
	The first column contains the year of the respective datapoint and the other
	columns contain the corresponding parameter values. The textfile needs to be tab 
	separated. The header of textfile has to start with `#` and needs to contain the tab separated
	column names (first being `Year` and the other being the same parameter
	names specified in the `Parameters - Monte_Carlo_Analysis` table.)
	If the textfile does not contain values for all parameters from the Monte Carlo analysis,
	missing parameter values are set to the base/reference value are are assumed to be 
	constant across all years included in the textfile.
	'''

	def __init__(self, input_file):
		self.inp = convert_input_to_dictionary(input_file)
		self.monte_carlo = Monte_Carlo_Analysis(input_file)
		self.mc_parameters = deepcopy(self.monte_carlo.parameters)

		if 'Display Parameters' in self.inp:
			self.color = self.inp['Display Parameters']['Color']['Value']
			self.display_name = self.inp['Display Parameters']['Name']['Value']
		else:
			self.color = 'darkgreen'
			self.display_name = 'Model'

		if self.inp['Development_Distance_Time_Analysis']['Log Normalization']['Value'] == 'False':
			self.log_normalize = False
		else:
			self.log_normalize = True

		self.base_year = self.inp['Development_Distance_Time_Analysis']['Base Year']['Value']
		self.limit_year = self.inp['Development_Distance_Time_Analysis']['Extrapolation Limit Year']['Value']

		self.import_data(self.inp['Development_Distance_Time_Analysis']['Input File']['Value'])
		self.map_parameters()
		self.harmonize_monte_carlo_parameters()
		self.historical_development_distance()
		self.fit_historical_development_distance()

	def import_data(self, file_name):
		'''Importing historical data from textfile.
		'''

		self.data = read_textfile(file_name, delimiter = '	')
		self.years = self.data[:,0]

		file_read = file_import(file_name, mode = 'r')

		for row_counter, line in enumerate(file_read):
			if line[0] != '#':
				break
			else:
				line_clean = line.strip(' #\n')
				column_names = parse_parameter(line_clean, delimiter = '	')

		self.column_names = column_names

	def map_parameters(self):
		'''Mapping parameters from historical data input file to Monte Carlo
		parameters.
		'''

		parameters = {}

		mapped = [False] * len(self.column_names)
		mapped[0] = True # Time column at index 0
		
		for counter, name in enumerate(self.column_names):
			if name in self.mc_parameters:
				parameters[name] = merge(self.mc_parameters[name], {'Index' : counter})
				mapped[counter] = True

		if False in mapped:
			unmapped_columns = np.asarray(self.column_names)[~np.asarray(mapped)]
			raise KeyError(f'Column(s) {unmapped_columns} could not be mapped to Monte Carlo parameters.')

		self.parameters_time = parameters

	def harmonize_monte_carlo_parameters(self):
		'''Checking if all Monte Carlo parameters are included in historical dataset.

		Notes
		-----
		If Monte Carlo parameters are missing, they are added with their
		historical value being set to the present reference/base value.
		'''

		max_column_index = len(self.column_names)

		for name, mc_parameter in self.mc_parameters.items():
			if name not in self.parameters_time:
				self.parameters_time[name] = mc_parameter
				self.parameters_time[name]['Index'] = max_column_index
				self.data = np.c_[self.data, np.ones(len(self.data)) * mc_parameter['Reference']]

				max_column_index += 1

				print(f'{__name__} is adding {name} (set to reference value) to historical development distance analysis parameters.')

		assert sorted(list(self.parameters_time)) == sorted(list(self.mc_parameters)), 'Monte Carlo parameters differ from historical parameter data.'

	def historical_development_distance(self):
		'''Calculating development distance for historical and Monte Carlo analysis data.

		Notes
		-----
		To ensure that historical development distances are consistent with the Monte
		Carlo derived development, Monte Carlo distances are recalculated using the 
		same hyperparameters. If `log_normalize` is set to `True`, all distances
		are calculated using logarithmic normalization.
		'''

		self.distances = calculate_distance(self.data, self.parameters_time, self.monte_carlo.principal,
											metric = 'cityblock', sum_distance = True,
											log_normalize = self.log_normalize)

		self.monte_carlo.development_distance(log_normalize = self.log_normalize,
											  sum_distance = True)

		self.monte_carlo.full_distance_cost_relationship(log_normalize = self.log_normalize,
														 sum_distance = True)

	def fit_historical_development_distance(self):
		'''Linear and asymptotic models are fitted to historical distances and are used 
		to extrapolate into the future.

		Notes
		-----
		To simply fitting of the exponential/asymptotic model, year values and distances are
		transformed to start at 0 and converge to 0, respectively. Both the linear and the
		asymptotic model are then fitted to the transformed data. The results are transformed
		back for visualization.
		'''

		years_shifted = self.years - np.amin(self.years)
		distances_shifted = -self.distances + 1

		p_expo = fit_generic(years_shifted, distances_shifted, exponential_decline, p_guess = np.ones(2))
		p_linear = fit_generic(years_shifted, distances_shifted, linear, p_guess = np.ones(2))

		years_shifted_extended = np.arange(0, self.limit_year - np.amin(self.years) + 1, 1)
		years_extended = years_shifted_extended + np.amin(self.years)

		self.expo_distance_model = -exponential_decline(p_expo, years_shifted_extended) + 1
		self.linear_distance_model = -linear(p_linear, years_shifted_extended) + 1

		self.years_extended = years_extended

		self.p_expo = p_expo
		self.p_linear = p_linear

	def determine_distance_time_correspondence(self, years, model, 
											   spacing = 1,
											   minimum_tick_distance = 0.1,
											   spacing_distances = np.array([1, 2, 3, 5, 10, 15, 20])):
		'''Mapping development distance to time for plotting.

		Parameters
		----------
		years : ndarray
			1D array of years, for which development distances have been computed
			using the specified model.
		model : ndarray
			1D array of modelled, time-dependent development distance values.
		spacing : int, optional
			Spacing of year ticks for visualization. Unit is years.
		minimum_tick_distance : float, optional
			Minimum allowable distance between year ticks in axis coordinates (0 to 1).
		spacing_distances : ndarray, optional
			Possible spacings which are to be used if specified spacing violates 
			`minimum_tick_distance`.

		Returns
		-------
		labels : ndarray
			Array of year tick labels.
		distances : ndarray
			Array of distances, to which year tick labels correspond.

		Notes
		-----
		If specified spacing leads to tick separation which is smaller than
		`minimum_tick_distance`, the spacing is increased by iterating through
		`spacing_distances` until the `minimum_tick_distance` criterion is fullfilled.
		'''

		def generate_distances(max_idx, label_spacing):
			labels = np.arange(self.base_year, years[max_idx]+1, label_spacing)
			distances_idx = find_nearest(years, labels)
			distances = model[distances_idx]
			distances[0] = 0

			return labels, distances

		max_idx = np.where(model <= 1)[0][-1]
		labels, distances = generate_distances(max_idx, spacing)

		spacing_idx = 0

		while np.amin(np.diff(distances)) < minimum_tick_distance and spacing_idx < len(spacing_distances):
			labels, distances = generate_distances(max_idx, spacing_distances[spacing_idx])
			spacing_idx += 1

		return labels.astype('int'), distances

	def generate_time_axis(self, ax, position, distances, labels, axis_label, color):
		'''Generating additional x axis for plot which maps development distance to time.
		'''

		ax.xaxis.set_ticks_position('bottom')
		ax.xaxis.set_label_position('bottom')
		ax.spines['bottom'].set_position(('axes', position))

		ax.set_xticks(distances)
		ax.set_xticklabels(labels)
		ax.set_xlabel(axis_label)

		ax.xaxis.label.set_color(color)
		ax.tick_params(axis = 'x', colors = color)
		ax.spines['bottom'].set_color(color)

	def plot_distance_histogram(self, ax = None, figure_lean = True,
							    linear_axis_y_pos = -0.4,
								expo_axis_y_pos = -0.8,
								linear_axis_label = 'Year (linear model)',
								expo_axis_label = 'Year (asymptotic model)',
								label_kwargs = {},
								hist_kwargs = {}, table_kwargs = {}, 
								image_kwargs = {}, plot_kwargs = {},  
								**kwargs):
		'''Plot distance histogram with additional axis mapping development distance
		to time.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		linear_axis_y_pos : float, optional
			y position of linear model time axis in axis coordinates.
		expo_axis_y_pos : float, optional
			y position of exponential/asymptotic model time axis in axis coordinates.
		linear_axis_label : str, optional
			String for label of linear model time axis.
		expo_axis_label : str, optional
			String for label of exponential/asymptotic model time axis.
		label_kwargs : dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Analysis.Development_Distance_Time_Analysis.Development_Distance_Time_Analysis.determine_distance_time_correspondence`.
		hist_kwargs: dict, optional
			Dictionary containg optional keyword arguments for
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.plot_distance_histogram`
		table_kwargs : dict, optional
			Dictionary containing optional keyword arguments for 
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.render_parameter_table`
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.
		'''


		kwargs = {**{'left': 0.2, 'right': 0.55, 'bottom': 0.5, 'top': 0.9, 'hspace': 0.2,
	 			     'fig_width': 10, 'fig_height': 2.5, 'font_size': 12,
	 			     'fig_height': 3.5,
	 			     'name': 'Distance_Time_Histogram'}, 
	 			  **kwargs, **plot_kwargs}

		table_kwargs = {**{'colWidths': [0.55, 0.25, 0.07, 0.25], 'format_cutoff': 7}, 
						**table_kwargs}

		image_kwargs = {**{'path': None, 'x': -0.35, 'y': 0.5, 'zoom': 0.08}, 
						**image_kwargs}

		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax

		cm = plt.get_cmap('plasma')

		self.monte_carlo.plot_distance_histogram(ax = ax, figure_lean = False,
												 xlabel = True,
												 table_kwargs = table_kwargs,
												 image_kwargs = image_kwargs,
											  	 **hist_kwargs)

		labels_linear, distances_linear = self.determine_distance_time_correspondence(
																self.years_extended, 
																self.linear_distance_model,
																**label_kwargs)
		labels_expo, distances_expo = self.determine_distance_time_correspondence(
																self.years_extended, 
																self.expo_distance_model,
																**label_kwargs)

		ax1 = ax.twiny()
		ax2 = ax.twiny()

		self.generate_time_axis(ax1, linear_axis_y_pos, distances_linear, labels_linear, 
								linear_axis_label, cm(0.25))
		self.generate_time_axis(ax2, expo_axis_y_pos, distances_expo, labels_expo, 
								expo_axis_label, cm(0.0))

		if figure_lean is True:
			figure.execute()
			return figure.fig

	def plot_distance_cost_relationship(self, ax = None, figure_lean = True,
										linear_axis_y_pos = -0.25,
										expo_axis_y_pos = -0.5,
										linear_axis_label = 'Year (linear model)',
										expo_axis_label = 'Year (asymptotic model)',
										label_kwargs = {},
										dist_kwargs = {}, table_kwargs = {}, 
										image_kwargs = {}, plot_kwargs = {}, 
										**kwargs):
		'''Plot distance/cost relationship with additional axis mapping development distance
		to time.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		linear_axis_y_pos : float, optional
			y position of linear model time axis in axis coordinates.
		expo_axis_y_pos : float, optional
			y position of exponential/asymptotic time axis in axis coordinates.
		linear_axis_label : str, optional
			String for label of linear model time axis.
		expo_axis_label : str, optional
			String for label of exponential/asymptotic model time axis.
		label_kwargs : dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Analysis.Development_Distance_Time_Analysis.Development_Distance_Time_Analysis.determine_distance_time_correspondence`.
		dist_kwargs: dict, optional
			Dictionary containg optional keyword arguments for
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.plot_distance_cost_relationship`
		table_kwargs : dict, optional
			Dictionary containing optional keyword arguments for 
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.render_parameter_table`
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.
		'''

		kwargs = {**{'left': 0.09, 'right': 0.5, 'bottom': 0.4, 'top': 0.95,
					 'fig_width': 9, 'fig_height': 4.5, 'font_size': 12,
					 'name': 'Distance_Cost_Time_Relationship'}, 
					 **kwargs, **plot_kwargs}

		dist_kwargs = {**{'log_scale': False, 'legend_loc': 'upper right'},
					   **dist_kwargs}

		table_kwargs = {**{'xpos': 1.05, 'ypos': 0.5, 'height': 0.5, 'edge_padding': 0.0}, 
						**table_kwargs}

		image_kwargs = {**{'path': None, 'x': 1.6, 'y': 0.2, 'zoom': 0.095}, 
						**image_kwargs}

		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax

		cm = plt.get_cmap('plasma')

		self.monte_carlo.plot_distance_cost_relationship(ax = ax, figure_lean = False,
														 xlim = [0, 1],
														table_kwargs = table_kwargs,
														image_kwargs = image_kwargs,
														**dist_kwargs)
		labels_linear, distances_linear = self.determine_distance_time_correspondence(
																self.years_extended, 
																self.linear_distance_model,
																**label_kwargs)
		labels_expo, distances_expo = self.determine_distance_time_correspondence(
																self.years_extended, 
																self.expo_distance_model,
																**label_kwargs)

		ax1 = ax.twiny()
		ax2 = ax.twiny()

		self.generate_time_axis(ax1, linear_axis_y_pos, distances_linear, labels_linear, 
								linear_axis_label, cm(0.25))
		self.generate_time_axis(ax2, expo_axis_y_pos, distances_expo, labels_expo, 
								expo_axis_label, cm(0.0))

		if figure_lean is True:
			figure.execute()
			return figure.fig

	def plot_distance_time_relationship(self, ax = None, figure_lean = True,  
										legend_loc = 'upper left',
										xlabel_string = 'Year',
										ylabel_string = 'Development distance',
										expo_label_string = 'Asymptotic model',
										linear_label_string = 'Linear model',
										datapoint_label_string = ' historical distance',
										markersize = 10,
										parameter_table = True,
										table_kwargs = {}, image_kwargs = {}, plot_kwargs = {},
										**kwargs):
		'''Ploting relationship between time and development distance based on 
		historical data.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		legend_loc : str, optional
			Controls location of legend in plot. Defaults to 'upper left'.
		xlabel_string : str, optional
			String for x axis label.
		ylabel_string : str, optional
			String for y axis label.
		expo_label_string : str, optional
			String for label of exponential/asymptotic model.
		linear_label_string : str, optional
			String for label of linear model.
		datapoint_label_string : str, optional
			Second part of string for label of historical datapoints. The first part
			is the display name of the model.
		markersize : float, optional
			Size of markers in scatter plot.
		parameter_table : bool, optional
			If parameter_table is True, the parameter table is shown in the plot.
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.		
		'''

		kwargs = {**{'left': 0.09, 'right': 0.5, 'bottom': 0.15, 'top': 0.95,
				     'fig_width': 9, 'fig_height': 4, 'font_size': 12,
				     'name': 'Distance_Time_Relationship'}, 
				      **kwargs, **plot_kwargs}

		table_kwargs = {**{'xpos': 1.05, 'ypos': 0.5, 'height': 0.5, 'edge_padding': 0.0}, 
						**table_kwargs}

		image_kwargs = {**{'path': None, 'x': 1.6, 'y': 0.2, 'zoom': 0.095}, 
						**image_kwargs}

		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax

		cm = plt.get_cmap('plasma')

		ax.plot(self.years_extended, self.linear_distance_model, color = cm(0.25),
				label = linear_label_string)
		ax.plot(self.years_extended, self.expo_distance_model, color = cm(0.0), 
				label = expo_label_string)
		ax.plot(self.years, self.distances, '.', color = self.color, markersize = markersize,
				label = self.display_name + datapoint_label_string)

		ax.axhspan(0, 1, color = cm(0.5), alpha = 0.1)
		ax.axvspan(self.base_year, np.amax(self.years_extended), color = cm(0.75), alpha = 0.1)

		ax.set_xlim(right = np.amax(self.years_extended))

		ax.set_xlabel(xlabel_string)
		ax.set_ylabel(ylabel_string)
		ax.legend(loc = legend_loc)

		ax.grid(color = 'grey', linestyle = '--', linewidth = 0.2, zorder = 0)

		if parameter_table is True:
			self.monte_carlo.render_parameter_table(ax, **table_kwargs)

		if image_kwargs['path'] is not None:
			insert_image(ax = ax, **image_kwargs)

		if figure_lean is True:
			figure.execute()
			return figure.fig

